@@ -2,10 +2,7 @@
 
 # Required imports
 import os.path
-try:
-    import pwd
-except ImportError:
-    pwd = None
+from getpass import getuser
 import datetime
 import re
 import struct
@@ -27,6 +24,17 @@ byte_ordering = {'b': '>', 'l': '<'}
 
 header_finder = re.compile(b'\*\n\*[\w, \d]+\n')
 spec_finder = re.compile(b'#\w+;\d+;\d+;[\w, \d\-]+;\d+\n')
+tracking_conversion = {
+    'File Header': 0,
+    'General Info': 0,
+    'PPF Attributes': 1,
+    'PPF Base Vectors': 2,
+    'Profiles': 3,
+    'Traces': -1,
+    '': -1
+}
+tracking_tags = [' in file header', ' in provenance section', ' in base vector section', ' in time slice', '']
+
 
 def pairwise(iterable):
     "s -> (s0,s1), (s1,s2), (s2, s3), ..."
@@ -34,19 +42,18 @@ def pairwise(iterable):
     next(b, None)
     return zip(a, b)
 
+
 def decode_spec(spec):
     """
     A spec line has the following format:
 
     #<dtype>;<no. of time steps (separated by newline)>;<no. of radial points (no separation character)>;<quantity name>;<no. of lines of metadata (always before binary)>
     """
-
     if not spec.startswith('#'):
         raise Exception(spec, ' is not a spec line')
     splitted = spec[1:-1].split(';')
     if len(splitted) != 5:
         raise Exception(spec, ' is not a spec line')
-
     jet_format = splitted[0]
     ignored = splitted[1]
     npoints = int(splitted[2])
@@ -54,31 +61,11 @@ def decode_spec(spec):
     nlines = int(splitted[4])
     return label, jet_format, npoints, nlines
 
-def decode_section(section, sec_num, endianness='>'):
-    m = header_finder.match(section)
-    header_string = section[:m.end()]
-    section_string = section[m.end():]
-
-    # Look for all data specs in the section (starts with #dtype;....)
-    spec_starts = [match.start() for match in spec_finder.finditer(section_string)]
-    spec_starts.append(len(section_string)) # To also read the last block
-    blocks = [section_string[start:next_start] for start, next_start in pairwise(spec_starts)]
-    if b''.join(blocks) != section_string:
-        raise Exception('Something weird happened, did not split all blocks correctly')
-
-    section_data = collections.OrderedDict()
-    section_info = collections.OrderedDict()
-    for block_num, block in enumerate(blocks):
-        var_label, block_data, var_info = decode_block(block)
-        section_data[var_label] = block_data
-        section_info[var_label] = var_info
-        section_info[var_label]['SECNUM'] = sec_num + 1
-
-    section_header = header_string[3:-1].decode('latin-1')
-    return section_header, section_data, section_info
-
 
 def decode_metadata(raw_metadata):
+    """
+    Metadata lines follow the spec line, typically consisting of 4 lines with additional lines for PPF provenance
+    """
     itag = None
     tags = ['UNITS', 'DESC', 'SCSTR', 'XBASE', 'UID', 'DDA', 'DTYPE', 'SEQ']
     meta = list(map(lambda x: x.decode('latin-1').strip(), raw_metadata))
@@ -90,23 +77,52 @@ def decode_metadata(raw_metadata):
             metadata['UNITS'] = None
     return metadata
 
-def decode_block(block, endianness='>'):
+
+def decode_block(block, endianness='>', track_num=-1):
+    """
+    A block is the entire portion between spec lines, and typically contain the data for a single variable
+    """
+    loc = tracking_tags[track_num] if track_num < 3 else tracking_tags[3]
+    if track_num >= 3:
+        loc += ' number {:d}'.format(track_num - 3)
+    corrupted = False
+    # Read specification
     m = spec_finder.match(block)
     spec_string = block[:m.end()]
     block_string = block[m.end():]
     spec = spec_string.decode('latin-1')
     var_label, jet_format, npoints, nlines = decode_spec(spec)
     var_format = dtypes[jet_format]
+    spec = spec.strip()
     # Read metadata
-    splitted = re.split(b'(\n)', block_string)
-    raw_metadata = splitted[:2*nlines:2]
-    raw_data = b''.join(splitted[2*nlines:-2])
-    var_metadata = decode_metadata(raw_metadata)
-    var_metadata['LABEL'] = var_label
-    var_metadata['FORM'] = jet_format
+    try:
+        splitted = re.split(b'(\n)', block_string)
+        raw_metadata = splitted[:2*nlines:2]
+        raw_data = b''.join(splitted[2*nlines:-2])
+    # Protection against invalid specification line
+    except Exception:
+        corrupted = True
+        raw_metadata = None
+        raw_data = None
+        print('Could not split {!s} block with spec {!s}{!s}'.format(var_label, spec, loc))
+    try:
+        var_metadata = decode_metadata(raw_metadata)
+        var_metadata['LABEL'] = var_label
+        var_metadata['FORM'] = jet_format
+    # Protection against incorrect number of metadata lines
+    except Exception:
+        corrupted = True
+        var_metadata = {}
+        print('Could not decode {!s} metadata with spec {!s}{!s}'.format(var_label, spec, loc))
+    # Read data
     if var_format is None:
         # No decoding needed, just read
-        decoded_var = raw_data.decode('latin-1').strip()
+        try:
+            decoded_var = raw_data.decode('latin-1').strip()
+        # Protection against plain read failure (highly unlikely)
+        except Exception:
+            corrupted = True
+            decoded_var = None
     else:
         if len(raw_data) == 0:
             # Empty variable
@@ -120,14 +136,64 @@ def decode_block(block, endianness='>'):
                     decoded_var = decoded_var[0]
                 elif var_format in ['i', 'f', 'd']:
                     decoded_var = np.array(decoded_var)
-            except:
-                raise Exception('Could not decode {!s} with spec {!s}'.format(var_label, spec))
+            # Protection against incorrect specification of binary line length / format
+            except Exception:
+                corrupted = True
+                decoded_var = np.full(npoints, np.nan)
+                print('Could not decode {!s} data with spec {!s}{!s}'.format(var_label, spec, loc))
+    # Scale numerical data based on metadata value
     if 'SCALE' in var_metadata:
         try:
             decoded_var *= var_metadata['SCALE']
-        except:
-            raise Exception('Cannot rescale', var_label)
-    return var_label.upper(), decoded_var, var_metadata
+        # Protection against invalid value in scaling factor metadata
+        except Exception:
+            corrupted = True
+            decoded_var = np.full(npoints, np.nan)
+            print('Could not rescale {!s}{!s}'.format(var_label, loc))
+    return var_label.upper(), decoded_var, var_metadata, corrupted
+
+
+def decode_section(section, sec_num, endianness='>', metadata=None, data_start_num=-1):
+    """
+    A section is the entire portion between lines beginning with **
+
+    Metadata option improves robustness of read routine by providing NaN vector shape in case of read failure
+    """
+    known_info = metadata if isinstance(metadata, dict) else {}
+    m = header_finder.match(section)
+    header_string = section[:m.end()]
+    section_string = section[m.end():]
+    section_header = header_string[3:-1].decode('latin-1')
+    # Look for all data specs in the section (starts with #dtype;....)
+    spec_starts = [match.start() for match in spec_finder.finditer(section_string)]
+    spec_starts.append(len(section_string)) # To also read the last block
+    blocks = [section_string[start:next_start] for start, next_start in pairwise(spec_starts)]
+    if b''.join(blocks) != section_string:
+        raise Exception('Something weird happened, did not split all blocks correctly')
+    section_data = collections.OrderedDict()
+    section_info = collections.OrderedDict()
+    track_num = tracking_conversion[section_header] if section_header in tracking_conversion else -1
+    if track_num == 3 and data_start_num > 0:
+        track_num += (sec_num - data_start_num)
+    for block_num, block in enumerate(blocks):
+        var_label, block_data, var_info, corrupted = decode_block(block, endianness=endianness, track_num=track_num)
+        record = True
+        # Check if variable metadata is already known (i.e. corrupted entry is not in the first time slice)
+        if corrupted and var_label in known_info:
+            var_info = metadata[var_label]
+        # Check if data already exists (i.e. duplicate entry in section)
+        if var_label in section_info and var_label in section_data:
+            record = False
+            # Check if existing data is already good, if yes, keep it
+            good_entry = ('XBASE' in section_info[var_label] and section_data[var_label] is not None and np.all(np.isfinite(section_data[var_label])))
+            if not good_entry and not corrupted:
+                record = True
+        if record:
+            section_data[var_label] = block_data
+            section_info[var_label] = var_info
+            section_info[var_label]['SECNUM'] = sec_num + 1
+    return section_header, section_data, section_info
+
 
 def convert_binary_file(input_file, output_file='binconv.txt'):
     """
@@ -135,6 +201,11 @@ def convert_binary_file(input_file, output_file='binconv.txt'):
     format as the original binary file. Effectively all binary
     numerical data is converted into ASCII. Can read .ex, .ext,
     .jsp, .jst, .jss, .jse files.
+
+    This function uses a different scheme from read_binary_file()
+    to read the binary data file, which is a bit slower. It can
+    be used to debug the canonical method under some conditions
+    and is kept for that reason.
     """
     supported_file_extensions = ['.ex', '.ext', '.jsp', '.jst', '.jss', '.jse']
     ipath = Path(input_file)
@@ -229,6 +300,7 @@ def convert_binary_file(input_file, output_file='binconv.txt'):
 
     return status
 
+
 def replace_none(any_dict, replace_val=[]):
     """
     Recursively replace None values in dict with replace_val.
@@ -240,7 +312,9 @@ def replace_none(any_dict, replace_val=[]):
         elif isinstance(v, (dict, collections.OrderedDict)):
             replace_none(v)
 
+
 def write_mat(binary_dict, output_file):
+    """ Saves input dictionary data in MATLAB .mat format """
     if savemat is not None:
         bf = copy.deepcopy(binary_dict)
         replace_none(bf)
@@ -248,6 +322,7 @@ def write_mat(binary_dict, output_file):
     else:
         raise IOError("MATLAB write requires scipy.io.savemat - please install scipy first.")
 
+
 def read_binary_file(input_path, output_file=None):
     """
     Reads a JETTO binary file into memory, for use within
@@ -287,10 +362,14 @@ def read_binary_file(input_path, output_file=None):
     section_header, section_data, header_info = decode_section(sections[0], 0)
     data[section_header] = section_data
     endianness = byte_ordering[section_data['FILE FORMAT']]
+    data_start_num = -1
 
     # Now decode all sections
     for sec_num, section in enumerate(sections[1:-1], start=1):
-        section_header, section_data, section_info = decode_section(section, sec_num, endianness=endianness)
+        metadata = data['INFO'] if data['INFO'] else None
+        if 'Profiles' in data and data_start_num < 0:
+           data_start_num = sec_num - 1
+        section_header, section_data, section_info = decode_section(section, sec_num, endianness=endianness, metadata=metadata, data_start_num=data_start_num)
         if section_header == 'Profiles':
             if 'Profiles' not in data:
                 data['Profiles'] = []
@@ -302,24 +381,50 @@ def read_binary_file(input_path, output_file=None):
             if k not in data['INFO']:
                  data['INFO'][k] = v
 
-    # Merge all profiles together into 2D arrays
+    # Check length of 1D profile arrays
     if 'Profiles' in data:
         profiles = collections.OrderedDict()
-        for prof in data['Profiles']:
+        for slice_num, prof in enumerate(data['Profiles']):
             for key in prof.keys():
                 if key not in profiles:
                     profiles[key] = []
-                profiles[key].append(prof[key])
-
+                prof_data = copy.deepcopy(prof[key])
+                if 'PPF Base Vectors' in data and 'INFO' in data and key in data['INFO'] and 'XBASE' in data['INFO'][key]:
+                    if not isinstance(prof_data, np.ndarray):
+                        prof_data = np.full(data['PPF Base Vectors'][data['INFO'][key]['XBASE']].size, np.nan)
+                    if prof_data.size != data['PPF Base Vectors'][data['INFO'][key]['XBASE']].size:
+                        prof_data = np.full(data['PPF Base Vectors'][data['INFO'][key]['XBASE']].size, np.nan)
+                if len(profiles[key]) != slice_num:
+                    fill_data = np.full(prof_data.shape, np.nan)
+                    profiles[key].append(fill_data)
+                    print('Missing {!s} data in time slice number {:d}'.format(key, slice_num - 1))
+                profiles[key].append(prof_data)
+        # Merge all 1D profiles of like quantities together into 2D arrays
         for key in list(profiles.keys()):
             profiles[key] = np.stack(profiles[key])
-
         data['Profiles'] = profiles
 
-    data = convert_data_to_EX2GK(data, header_info, endianness)
+    # Check length of 1D time arrays
+    if 'Traces' in data:
+        traces = collections.OrderedDict()
+        for key in data['Traces'].keys():
+            if key not in traces:
+                traces[key] = []
+            trac_data = copy.deepcopy(data['Traces'][key])
+            if 'PPF Base Vectors' in data and 'INFO' in data and key in data['INFO'] and 'XBASE' in data['INFO'][key]:
+                if not isinstance(trac_data, np.ndarray):
+                    trac_data = np.full(data['PPF Base Vectors'][data['INFO'][key]['XBASE']].size, np.nan)
+                if trac_data.size != data['PPF Base Vectors'][data['INFO'][key]['XBASE']].size:
+                    trac_data = np.full(data['PPF Base Vectors'][data['INFO'][key]['XBASE']].size, np.nan)
+            traces[key] = trac_data
+        data['Traces'] = traces
+
+    data = standardize_data_representation(data, header_info, endianness)
     return data
 
-def convert_data_to_EX2GK(data, header_info, endianness):
+
+def standardize_data_representation(data, header_info, endianness):
+    """ Standardize field formats and labels """
     data['SECTIONS'] = [key for key in data.keys() if key != 'INFO']
     for key in data['SECTIONS']:
         dateval = None
@@ -444,6 +549,11 @@ def create_exfile_structure(database,shot,version_tag=None,extfile=False):
     data['INFO']['USER PRE-MODEX EX-FILE'] = {'FORM': 'char', 'LABEL': 'User Pre-Modex EX-file', 'SECNUM': 2}
     data['INFO']['SHOT'] = {'FORM': 'int', 'LABEL': 'Shot', 'SECNUM': 3}
     data['INFO']['DDA NAME'] = {'FORM': 'char', 'LABEL': 'DDA Name', 'SECNUM': 3}
+    if not extfile:
+        data['INFO']['XVEC1'] = {'UNITS': None, 'DESC': 'RHO normalised', 'SCSTR': '1.0', 'SCALE': 1.0, 'LABEL': 'XVEC1', 'FORM': 'float', 'SECNUM': 4}
+        data['INFO']['TVEC1'] = {'UNITS': 'secs', 'DESC': 'TIME', 'SCSTR': '1.0', 'SCALE': 1.0, 'LABEL': 'TVEC1', 'FORM': 'float', 'SECNUM': 5}
+    else:
+        data['INFO']['TVEC1'] = {'UNITS': 'secs', 'DESC': 'TIME', 'SCSTR': '1.0', 'SCALE': 1.0, 'LABEL': 'TVEC1', 'FORM': 'float', 'SECNUM': 4}
 
     # Add required metadata - covers until end of 'PPF Attributes' section
     data['FILE FORMAT'] = 'b'
@@ -456,6 +566,11 @@ def create_exfile_structure(database,shot,version_tag=None,extfile=False):
     data['USER PRE-MODEX EX-FILE'] = ''
     data['SHOT'] = np.atleast_2d([int(shot)]) 
     data['DDA NAME'] = 'EX'
+    if not extfile:
+        data['XVEC1'] = np.atleast_2d([])
+        data['TVEC1'] = np.atleast_3d([])
+    else:
+        data['TVEC1'] = np.atleast_2d([])
 
     return data
 
@@ -480,10 +595,10 @@ def write_binary_exfile(data,output_file='custom'):
     ofname = 'custom.ex'
     if isinstance(output_file,str):
         ofname = output_file
-    if not ofname.endswith('.ex'):
+    if 'Profiles' in ds and not ofname.endswith('.ex'):
         ofname = ofname + '.ex'
-    if "TVEC1" in di and di["TVEC1"]["SECNUM"] == (baseidx + 1) and not ofname.endswith('.ext'):
-        ofname = ofname + 't'
+    if 'Traces' in ds and not ofname.endswith('.ext'):
+        ofname = ofname + '.ext' if not ofname.endswith('.ex') else ofname + 't'
     opath = Path(ofname)
     if not opath.parent.exists():
         opath.parent.mkdir(parents=True)
@@ -675,8 +790,6 @@ def add_entry(data,key,moddata,dtype,units,description,scale,xbase,tag=None):
         itag = itag + ' - ' + tag
 
     odata = copy.deepcopy(data)
-    if pwd is None:
-        raise Exception('pwd python module not loaded. Are you sure this is a Linux system?')
     if isinstance(data,(dict,collections.OrderedDict)) and ikey is not None:
         check_flag = True
         if ikey not in ["TVEC1", "XVEC1"]:
@@ -691,7 +804,7 @@ def add_entry(data,key,moddata,dtype,units,description,scale,xbase,tag=None):
             check_flag = all(check_array)
         if check_flag:
             if ikey not in data:
-                uname = pwd.getpwuid(os.getuid())[0]
+                uname = getuser()
                 odata[ikey] = idata.copy()
                 odata["INFO"][ikey] = dict()
                 odata["INFO"][ikey]["UNITS"] = units            # Units string
@@ -745,13 +858,11 @@ def modify_entry(data,key,moddata,tag=None):
         itag = itag + ' - ' + tag
 
     odata = None
-    if pwd is None:
-        raise Exception('pwd python module not loaded. Are you sure this is a Linux system?')
     if isinstance(data,(dict,collections.OrderedDict)) and ikey in data:
         if data[ikey].size == idata.size:
             idata = np.reshape(idata,data[ikey].shape)
         if data[ikey].shape == idata.shape:
-            uname = pwd.getpwuid(os.getuid())[0]
+            uname = getuser()
             odata = copy.deepcopy(data)
             odata[ikey] = idata.copy()
             odata["INFO"][ikey]["UID"] = uname    # User ID (auto-detected)
