@@ -131,6 +131,26 @@ class Template:
         """
         return param['jset_id'] is not None
 
+    @classmethod
+    def extract_jset_ids(cls, param: Dict) -> List[str]:
+        """Extract jset_ids, as they can be zero, one or many per parameter
+        
+        Returns a tuple with those lists and the dimension
+
+        jset_ids, namelist_ids, fields, dimension = extract_all_fields(param)
+
+        :param param: Parameter specification
+        :type param: Dict
+        """
+
+        jset_ids = []
+        if 'jset_id' in param.keys():
+            jset_ids.append(param['jset_id'])
+        if 'jset_flex_id' in param.keys():
+            jset_ids += param['jset_flex_id']
+
+        return jset_ids
+
     @classmethod
     def validate_regular_param(cls, name: str, param: Dict, jset, namelist: jetto_tools.namelist.Namelist,
                                sanco_namelist: Union[None, jetto_tools.namelist.Namelist]):
@@ -151,16 +171,20 @@ class Template:
         :type sanco_namelist: Union[None, jetto_tools.namelist.Namelist]
         :raise: TemplateError if any of the checks fail
         """
-        jset_id = param['jset_id']
-        namelist_id = param['nml_id']['namelist']
-        field = param['nml_id']['field']
+        jset_ids = Template.extract_jset_ids(param)
         dimension = param['dimension']
 
         if dimension == 'vector':
             raise TemplateError(f'Vector parameter {name} not supported outside of extra namelists')
+        for jset_id in jset_ids:
+            if jset_id not in jset:
+                raise TemplateError(f'jset_id {jset_id} not found in template JSET')
 
-        if jset_id not in jset:
-            raise TemplateError(f'jset_id {jset_id} not found in template JSET')
+        if 'nml_id' not in param:
+            return
+
+        namelist_id = param['nml_id']['namelist']
+        field = param['nml_id']['field']
 
         if not namelist.exists(namelist_id, field):
             if sanco_namelist is None or not sanco_namelist.exists(namelist_id, field):
@@ -192,7 +216,7 @@ class Template:
             raise TemplateError(f'Extra namelist parameter {name} not found'
                                 ' in JETTO/SANCO extra namelists')
 
-        data_type = {'int': int, 'real': float}[param['type']]
+        data_type = {'int': int, 'real': float, 'str': str}[param['type']]
         dimension = param['dimension']
         if field in jset.extras:
             cls.validate_extra_namelist_param_details(field, jset.extras, namelist, data_type, 'JETTO')
